using System;
using System.Collections.Generic;
using System.Linq;

using GraphVertex = System.Int32;

class GraphValidations
{		
	static void Main() 
	{
		Console.WriteLine("\n\n");
		
		Graph graph1 = new Graph("Graph 1", Edge(1, 2), Edge(1, 3), Edge(4, 5), Edge(6, 7), Edge(6, 8), Edge(6, 9), Edge(10, 11));
		Graph graph2 = new Graph("Graph 2", Edge(1, 2), Edge(2, 3), Edge(3, 4), Edge(4, 5), Edge(1, 3), Edge(1, 4), Edge(1, 5), Edge(2, 4), Edge(2, 5), Edge(3, 5));
		Graph graph3 = new Graph("Graph 3", Edge(1, 2), Edge(2, 3), Edge(3, 4), Edge(1, 3), Edge(1, 4), Edge(5, 6), Edge(5, 7));

		IsTransitiveProcedural(graph1);
		IsTransitiveProcedural(graph2);
		IsTransitiveProcedural(graph3);

		Console.WriteLine("\n\n");

		IsTransitiveLogicStyle(graph1);
		IsTransitiveLogicStyle(graph2);
		IsTransitiveLogicStyle(graph3);
		
		Console.WriteLine("\n\n");
		
		Graph graph4 = new Graph("Graph 4", Edge(1, 6), Edge(2, 3), Edge(3, 4), Edge(4, 5), Edge(5, 6), Edge(6, 1), Edge(6, 2), Edge(6, 3), Edge(6, 4), Edge(6, 5), Edge(6, 6));
		Graph graph5 = new Graph("Graph 5", Edge(1, 1), Edge(1, 2), Edge(1, 3), Edge(1, 4), Edge(2, 3), Edge(3, 1), Edge(3, 2), Edge(3, 3), Edge(3, 4), Edge(4, 1));
		Graph graph6 = new Graph("Graph 6", Edge(1, 1), Edge(1, 2), Edge(1, 3), Edge(1, 4), Edge(3, 1), Edge(3, 2), Edge(3, 3), Edge(3, 4), Edge(4, 1));
		Graph graph7 = new Graph("Graph 7", Edge(1, 1), Edge(1, 3), Edge(1, 4), Edge(1, 5), Edge(2, 1), Edge(3, 2), Edge(3, 3), Edge(3, 4), Edge(3, 5), Edge(4, 5), Edge(5, 1));
		
		DoAllVerticesAccessCentralVertexProcedural(graph4);
		DoAllVerticesAccessCentralVertexProcedural(graph5);
		DoAllVerticesAccessCentralVertexProcedural(graph6);
		DoAllVerticesAccessCentralVertexProcedural(graph7);
		
		Console.WriteLine("\n\n");
		
		DoAllVerticesAccessCentralVertexLogicStyle(graph4);
		DoAllVerticesAccessCentralVertexLogicStyle(graph5);
		DoAllVerticesAccessCentralVertexLogicStyle(graph6);
		DoAllVerticesAccessCentralVertexLogicStyle(graph7);
		
		Console.WriteLine("\n\nPress ENTER to exit.");
		Console.ReadKey();
	}

	// A method that tests whether a given graph is transitive. Uses the procedural (old) style verification.
	static void IsTransitiveProcedural(Graph graph) =>
		Console.WriteLine("Is graph \"{0}\" transitive? Answer: {1}.\n", graph.Name, graph.IsTransitiveProcedural());

	// A method that tests whether a given graph is transitive. Uses the logic programming (new) style verification.
	static void IsTransitiveLogicStyle(Graph graph) =>
		Console.WriteLine("Is graph \"{0}\" transitive? Answer: {1}.\n", graph.Name, graph.IsTransitiveLogicStyle());

	// A method that tests whether in a given graph every vertex is connected to a central vertex. A central vertex is one edge away from all vertices in the graph .Uses the procedural (old) style verification.
	static void DoAllVerticesAccessCentralVertexProcedural(Graph graph) =>
		Console.WriteLine("Is every vertex in graph \"{0}\" connected to a central vertex? Answer: {1}.\n", graph.Name, graph.DoAllVerticesAccessCentralVertexProcedural());
	
	// A method that tests whether in a given graph every vertex is connected to a central vertex. A central vertex is one edge away from all vertices in the graph .Uses the logic programming (new) style verification.
	static void DoAllVerticesAccessCentralVertexLogicStyle(Graph graph) =>
		Console.WriteLine("Is every vertex in graph \"{0}\" connected to a central vertex? Answer: {1}.\n", graph.Name, graph.DoAllVerticesAccessCentralVertexLogicStyle());

	class GraphEdge : Tuple<GraphVertex, GraphVertex>
	{
		public GraphEdge(GraphVertex fromVertex, GraphVertex toVertex) : base(fromVertex, toVertex)	{}
		
		public GraphVertex From => Item1;
		public GraphVertex To => Item2;
	}
	
	// Syntactic sugar meant to ease the declaration of edges
	static GraphEdge Edge(GraphVertex inVertex, GraphVertex outVertex)
	{
		return new GraphEdge(inVertex, outVertex);
	}
	
	class Graph
	{
		public Graph(string name, params GraphEdge[] edges)
		{
			Name = name;
			foreach(var edge in edges)	AddEdge(edge);
		}
		
		private void AddEdge(GraphEdge edge)
		{
			vertices.Add(edge.From);
			vertices.Add(edge.To);
			edges.Add(edge);
		}
		
		// Verifies whether two vertices are connected via a path
		private bool AreConnected(GraphVertex startVertex, GraphVertex endVertex) => 
			DFSTraversal(startVertex).Contains(endVertex);
		
		// Performs a depth-first-search traversal of the graph starting from the given vertex
		private IEnumerable<GraphVertex> DFSTraversal(GraphVertex startVertex) => 
			DFSTraversal(startVertex, new HashSet<GraphVertex>());
		
		// Performs a depth-first-search traversal of the graph starting from the given vertex. A collection of visited vertices is maintained to prevent infinite loop.
		private IEnumerable<GraphVertex> DFSTraversal(GraphVertex startVertex, HashSet<GraphVertex> visitedVertices)
		{
			yield return startVertex;
			visitedVertices.Add(startVertex);
			
			foreach(var successor in SuccessorsOf(startVertex).Where(successor => !visitedVertices.Contains(successor)))
				foreach(var vertex in DFSTraversal(successor, visitedVertices))
					yield return vertex;
		}
		
		// Collects all vertices that are accessed by a single edge from the given vertex
		private IEnumerable<GraphVertex> SuccessorsOf(GraphVertex startVertex) => 
			vertices.Where(vertex => edges.Contains(Edge(startVertex, vertex)));
		
		public string Name {	get;	}
		private HashSet<GraphEdge> edges = new HashSet<GraphEdge>();
		private HashSet<GraphVertex> vertices = new HashSet<GraphVertex>();
		
		// Verifies whether the graph is transitive. The verification is done in procedural (old) style.
		public bool IsTransitiveProcedural()
		{
			foreach(var x in vertices)	foreach(var y in vertices) foreach(var z in vertices)
				if(edges.Contains(Edge(x, y)) && edges.Contains(Edge(y, z)) && !edges.Contains(Edge(x, z))) return false;
			
			return true;
		}
		
		// Verifies whether the graph is transitive. The verification is done in logic programming (new) style.
		public bool IsTransitiveLogicStyle() => 
			! vertices.Any(x => vertices.Any(y => vertices.Any(z => 
				edges.Contains(Edge(x, y)) && edges.Contains(Edge(y, z)) && !edges.Contains(Edge(x, z)))));
				
		// Verifies whether every vertex is connected to a central vertex. A central vertex is one edge away from all vertices in the graph. The verification is done in procedural (old) style.
		public bool DoAllVerticesAccessCentralVertexProcedural()
		{
			foreach(var x in vertices)
			{
				bool found = false;
				
				foreach(var y in vertices)
				{
					if(AreConnected(x, y))
					{
						bool isYCentral = true;
						foreach(var z in vertices)
						{
							if(!edges.Contains(Edge(y, z)))
							{
								isYCentral = false;
								break;
							}
						}
						
						if(isYCentral)
						{
							found = true;
							break;
						}
					}
				}
				
				if(!found) return false;
			}
			
			return true;
		}
		
		// Verifies whether every vertex is connected to a central vertex. A central vertex is one edge away from all vertices in the graph. The verification is done in logic programming (new) style.
		public bool DoAllVerticesAccessCentralVertexLogicStyle() => 
			vertices.All(x => vertices.Any(y => AreConnected(x, y) && vertices.All(z => edges.Contains(Edge(y, z)))));
	}
}
