using System;
using System.Collections.Generic;
using System.Linq;

using GraphVertex = System.Int32;

class LINQQueries
{		
	static void Main() 
	{
		Console.WriteLine("\n\n");
			
		// Initialize data collections with sales and customers
		List<Customer> customers 	= 
			new List<Customer>(new Customer[] {
				new Customer(1, "Number one", false), 
				new Customer(2, "Data", true), 
				new Customer(3, "Counselor Troi", false), 
				new Customer(4, "Q", true), 
				new Customer(5, "Worf", false)
			});
		List<Sale> sales = 
			new List<Sale>(new Sale[] {
				new Sale(1, new DateTime(2016, 11, 4), 2, 2500), 
				new Sale(2, new DateTime(2015, 9, 21), 1, 1100), 
				new Sale(3, new DateTime(2015, 12, 17), 2, 2000), 
				new Sale(4, new DateTime(2016, 10, 5), 1, 1500), 
				new Sale(5, new DateTime(2016, 10, 5), 4, 5000), 
				new Sale(6, new DateTime(2016, 2, 18), 4, 750), 
				new Sale(7, new DateTime(2016, 7, 1), 2, 3000), 
				new Sale(8, new DateTime(2015, 11, 29), 3, 120), 
				new Sale(9, new DateTime(2015, 4, 15), 2, 5600), 
				new Sale(10, new DateTime(2015, 12, 10), 4, 1200)
			});
		
		// Use simple query to list the ids of all sales during year 2016
		var saleIds2016 = 
			from sale in sales
			where sale.Date.Year == 2016
			select sale.SaleId;
		saleIds2016.ForEach(saleId => Console.WriteLine("Id of a sale within year 2016 is: {0}.", saleId));
		
		Console.WriteLine("\n\n");
		
		// Retrieve id and date of all sales with amount greater than 2000
		var salesOver2000 = 
			from sale in sales
			where sale.Amount > 2000
			select new { sale.SaleId, sale.Date };
		salesOver2000.ForEach(data => Console.WriteLine("Sale with id {0} on {1} features amount over 2000.", data.SaleId, data.Date));
		
		Console.WriteLine("\n\n");
		
		// Use composite conditions about date and amount to search for sale ids
		var saleIds2016Over2000 = 
			from sale in sales
			where sale.Date.Year == 2016
			where sale.Amount > 2000
			select sale.SaleId;
		saleIds2016Over2000.ForEach(saleId => Console.WriteLine("Id of a sale within year 2016 that features amount over 2000 is: {0}.", saleId));
		
		Console.WriteLine("\n\n");
		
		// Retrieve corporate customer ids and names from sales with amount greater than 2000
		var customersOver2000 = 
			from sale in sales
				join customer in customers
					on sale.CustomerId equals customer.CustomerId
			where sale.Amount > 2000
			select new { sale.SaleId, customer.CustomerId, customer.CustomerName };
		customersOver2000.ForEach(data => Console.WriteLine("Customer data from sale {0} with amount over 2000 is: id = {1}, name = \"{2}\".", data.SaleId, data.CustomerId, data.CustomerName));
		
		Console.WriteLine("\n\n");
		
		// Order the sales by customer name (ascending) and then by date (descending)
		var ordered = 
			from sale in sales
				join customer in customers
					on sale.CustomerId equals customer.CustomerId
			orderby customer.CustomerName, sale.Date descending
			select new { sale.SaleId, customer.CustomerName, sale.Date };
		ordered.ForEach(data => Console.WriteLine("Sale {0} is for customer \"{1}\" on {2}.", data.SaleId, data.CustomerName, data.Date));		
		
		Console.WriteLine("\n\n");
		
		// Calculate total sales amounts for each customer
		var totalAmounts = 
			from sale in sales
			group sale by sale.CustomerId into total
			select new { CustomerId = total.Key, TotalAmount = total.Sum(s => s.Amount) };
		totalAmounts.ForEach(data => Console.WriteLine("The total amount for customer with id {0} is: {1}.", data.CustomerId, data.TotalAmount));		
		
		Console.WriteLine("\n\n");
		
		// Find the totals for all corporate customers with total amount over 10000 (first take)
		var totalCorporateOver10000I = 
			from sale in sales
				join customer in customers
					on sale.CustomerId equals customer.CustomerId
			where customer.CorporateCustomer
			group new { customer.CustomerId, customer.CustomerName, sale.Amount } by customer.CustomerId into total
			where total.Sum(x => x.Amount) > 10000
			select new { CustomerId = total.Key, CustomerName = total.Select(x => x.CustomerName).First(), TotalAmount = total.Sum(x => x.Amount) };
		totalCorporateOver10000I.ForEach(data => Console.WriteLine("The total amount (at least 10000) for corporate customer with id {0} and name \"{1}\" is: {2}.", data.CustomerId, data.CustomerName, data.TotalAmount));	
		
		Console.WriteLine("\n\n");
		
		// Find the totals for all corporate customers with total amount over 10000 (second take)
		var totalCorporateOver10000II = 
			from sale in sales
				join customer in customers
					on sale.CustomerId equals customer.CustomerId
			where customer.CorporateCustomer
			group new { customer.CustomerId, customer.CustomerName, sale.Amount } by customer.CustomerId into total
			select new { CustomerId = total.Key, CustomerName = total.Select(x => x.CustomerName).First(), TotalAmount = total.Sum(x => x.Amount) } 
				into TotalData
				where TotalData.TotalAmount > 10000
				orderby TotalData.CustomerName 
				select TotalData;
		totalCorporateOver10000II.ForEach(data => Console.WriteLine("The total amount (at least 10000) for corporate customer with id {0} and name \"{1}\" is: {2}.", data.CustomerId, data.CustomerName, data.TotalAmount));	
		
		Console.WriteLine("\n\nPress ENTER to exit.");
		Console.ReadKey();
	}

	class Sale
	{
		public Sale(int saleId, DateTime date, int customerId, decimal amount)
		{
			SaleId = saleId;
			Date = date;
			CustomerId = customerId;
			Amount = amount;
		}
		
		public int SaleId;
		public DateTime Date;
		public int CustomerId;
		public decimal Amount;
	}
	
	class Customer
	{
		public Customer(int customerId, string customerName, bool corporateCustomer)
		{
			CustomerId = customerId;
			CustomerName = customerName;
			CorporateCustomer = corporateCustomer;
		}
		
		public int CustomerId;
		public string CustomerName;
		public bool CorporateCustomer;
	}
}

// Extending the IEnumerable<T> interface with a ForEach method (like the one from List<T>)
public static class IEnumerableExtensions
{
	public static void ForEach<T>(this IEnumerable<T> iEnumerable, Action<T> action)
	{
		foreach(T item in iEnumerable)	action(item);
	}
}
