using System;
using System.Collections.Generic;
using System.Linq;

using GraphVertex = System.Int32;

class TopologicalSort
{		
	static void Main() 
	{
		Console.WriteLine("\n\n");
		
		Graph graph = new Graph(Edge(3, 11), Edge(5, 10), Edge(3, 8), Edge(5, 6), Edge(7, 8), Edge(1, 2), Edge(8, 9), Edge(2, 4), Edge(1, 3), Edge(6, 7), Edge(1, 4), Edge(2, 5), Edge(4, 5));
		graph.TopologicalSort().ForEach(vertex => Console.WriteLine("Consecutive vertex in the topological sort: " + vertex));
		
		Console.WriteLine("\n\nPress ENTER to exit.");
		Console.ReadKey();
	}

	class GraphEdge : Tuple<GraphVertex, GraphVertex>
	{
		public GraphEdge(GraphVertex fromVertex, GraphVertex toVertex) : base(fromVertex, toVertex)	{}
		
		public GraphVertex From => Item1;
		public GraphVertex To => Item2;
	}
	
	// Syntactic sugar meant to ease the declaration of edges
	static GraphEdge Edge(GraphVertex inVertex, GraphVertex outVertex)
	{
		return new GraphEdge(inVertex, outVertex);
	}
	
	class Graph
	{
		public Graph(params GraphEdge[] graphEdges)
		{
			edges = new HashSet<GraphEdge>(graphEdges);
			
			vertices = new HashSet<GraphVertex>(edges.Select(edge => edge.From));
			vertices.UnionWith(edges.Select(edge => edge.To));
		}
		
		private HashSet<GraphVertex> vertices;
		private HashSet<GraphEdge> edges;
		
		// Performs a topological sort of the vertices of the graph
		public IEnumerable<GraphVertex> TopologicalSort()
		{			
			Dictionary<GraphVertex, HashSet<GraphEdge>> incomming = 
				edges.GroupBy(edge => edge.To).
					ToDictionary(grouping => grouping.Key, grouping => new HashSet<GraphEdge>(grouping.ToList()));
			Dictionary<GraphVertex, HashSet<GraphEdge>> outgoing = 
				edges.GroupBy(edge => edge.From).
					ToDictionary(grouping => grouping.Key, grouping => new HashSet<GraphEdge>(grouping.ToList()));
			
			HashSet<GraphVertex> front = 
				new HashSet<GraphVertex>(vertices.Where(vertex => incomming.GetOrDefault(vertex).Count == 0));
			while(front.Count > 0)
			{
				GraphVertex current = front.First();
				yield return current;
				
				front.Remove(current);
				outgoing.GetOrDefault(current).
					Select(edge => edge.To).
						ForEach(vertex => incomming.GetOrDefault(vertex).Remove(Edge(current, vertex)));				
				front.UnionWith(
					outgoing.GetOrDefault(current).
						Select(edge => edge.To).
							Where(vertex => incomming.GetOrDefault(vertex).Count == 0));
			}
		}
	}
}

// Extending the IEnumerable<T> interface with a ForEach method (like the one from List<T>)
public static class IEnumerableExtensions
{
	public static void ForEach<T>(this IEnumerable<T> iEnumerable, Action<T> action)
	{
		foreach(T item in iEnumerable)	action(item);
	}
}

// Extending the IDictionary<TKey, HashSet<TValue>> interface with a GetOrDefault method (like the one from Java's Hashtable)
public static class IDictionaryExtensions
{
	public static HashSet<TValue> GetOrDefault<TKey, TValue>(this IDictionary<TKey, HashSet<TValue>> iDictionary, TKey key)
	{
		return iDictionary.ContainsKey(key) ? iDictionary[key] : new HashSet<TValue>();
	}
}
